"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


import copy
import functools
import inspect
import re
import unicodedata

from collections import OrderedDict
from typing import (TYPE_CHECKING, Any, List, Dict, Optional, Iterable,
                    Callable, Sequence, Union, Tuple)
from fortnitepy.typedefs import MaybeCoro
from fortnitepy.party import ClientParty
from fortnitepy.friend import Friend

from .core import Group, Command
from .errors import CommandError
from .context import Context
from .cog import Cog

if TYPE_CHECKING:
    from .bot import Bot


__all__ = (
    'Paginator',
    'HelpCommand',
    'FortniteHelpCommand',
)

_IS_ASCII = re.compile(r'^[\x00-\x7f]+$')


def _string_width(string: str, *, _IS_ASCII: Any = _IS_ASCII) -> int:
    """Returns string's width."""
    match = _IS_ASCII.match(string)
    if match:
        return match.endpos

    UNICODE_WIDE_CHAR_TYPE = 'WFA'
    width = 0
    func = unicodedata.east_asian_width
    for char in string:
        width += 2 if func(char) in UNICODE_WIDE_CHAR_TYPE else 1
    return width


async def maybe_coroutine(func: MaybeCoro,
                          *args: list,
                          **kwargs: dict) -> Any:
    value = func(*args, **kwargs)
    if inspect.isawaitable(value):
        return await value
    else:
        return value


class Paginator:
    """A class that aids in paginating code blocks for Fortnite messages.

    .. container:: operations

        .. describe:: len(x)
            Returns the total number of characters in the paginator.

    Attributes
    -----------
    prefix: :class:`str`
        The prefix inserted to every page.
    suffix: :class:`str`
        The suffix appended at the end of every page.
    max_size: :class:`int`
        The maximum amount of codepoints allowed in a page.
    """

    def __init__(self, prefix: str = '',
                 suffix: str = '',
                 max_size: int = 10000) -> None:
        self.prefix = prefix
        self.suffix = suffix
        self.max_size = max_size
        self.clear()

    def clear(self) -> None:
        """Clears the paginator to have no pages."""
        if self.prefix is not None:
            self._current_page = [self.prefix]
            self._count = len(self.prefix)
        else:
            self._current_page = []
            self._count = 0

        self._pages = []

    @property
    def _prefix_len(self) -> int:
        return len(self.prefix) if self.prefix else 0

    @property
    def _suffix_len(self) -> int:
        return len(self.suffix) if self.suffix else 0

    def add_page(self, text: str) -> None:
        """Adds a page to the paginator with no additional checks done."""
        self._pages.append(text)

    def add_line(self, line: str = '', *, empty: bool = False) -> None:
        """Adds a line to the current page.

        If the line exceeds the :attr:`max_size` then an exception
        is raised.

        Parameters
        -----------
        line: :class:`str`
            The line to add.
        empty: :class:`bool`
            Indicates if another empty line should be added.

        Raises
        ------
        RuntimeError
            The line was too big for the current :attr:`max_size`.
        """
        max_page_size = self.max_size - self._prefix_len - self._suffix_len
        if len(line) > max_page_size:
            raise RuntimeError('Line exceeds maximum page size '
                               '{}'.format(max_page_size))

        if self._count + len(line) + 1 > self.max_size - self._suffix_len:
            self.close_page()

        self._count += len(line) + 1
        self._current_page.append(line)

        if empty:
            self._current_page.append('')
            self._count += 1

    def close_page(self) -> None:
        """Prematurely terminate a page."""

        if self.suffix is not None:
            self._current_page.append(self.suffix)

        self._pages.append('\n'.join(self._current_page))

        if self.prefix is not None:
            self._current_page = []
            self._count = len(self.prefix)
        else:
            self._current_page = []
            self._count = 0

    def __len__(self) -> int:
        total = sum(len(p) for p in self._pages)
        return total + self._count

    @property
    def pages(self) -> List[str]:
        """Returns the rendered list of pages."""

        if len(self._current_page) > (0 if self.prefix is None else 1):
            self.close_page()

        return self._pages

    def __repr__(self) -> str:
        fmt = ('<Paginator prefix: {0.prefix} suffix: {0.suffix} max_size: '
               '{0.max_size} count: {0._count}>')
        return fmt.format(self)


def _not_overridden(func: MaybeCoro) -> MaybeCoro:
    func.__fnpy_help_command_not_overridden__ = True
    return func


class _HelpCommandImpl(Command):
    def __init__(self, inject: Command, *args: list, **kwargs: dict) -> None:
        super().__init__(inject.command_callback, *args, **kwargs)
        self._original = inject
        self._injected = inject

    async def prepare(self, ctx: Context) -> None:
        self._injected = injected = self._original.copy()
        injected.context = ctx
        self.callback = injected.command_callback

        error_handler = injected.help_command_error_handler
        if not hasattr(error_handler, '__fnpy_help_command_not_overridden__'):
            if self.cog is not None:
                self.error_handler = self._error_handler_cog_implementation
            else:
                self.error_handler = error_handler

        await super().prepare(ctx)

    async def _parse_arguments(self, ctx: Context) -> None:
        # Make the parser think we don't have a cog so it doesn't
        # inject the parameter into `ctx.args`.
        original_cog = self.cog
        self.cog = None
        try:
            await super()._parse_arguments(ctx)
        finally:
            self.cog = original_cog

    async def _error_handler_cog_implementation(self, _,
                                                ctx: Context,
                                                error: Exception) -> None:
        await self._injected.help_command_error_handler(ctx, error)

    @property
    def clean_params(self) -> OrderedDict:
        result = self.params.copy()
        try:
            result.popitem(last=False)
        except Exception:
            raise ValueError('Missing context parameter') from None
        else:
            return result

    def _inject_into_cog(self, cog: Cog) -> None:
        # Warning: hacky

        # Make the cog think that get_commands returns this command
        # as well if we inject it without modifying __cog_commands__
        # since that's used for the injection and ejection of cogs.
        def wrapped_get_commands(*, _original=cog.get_commands):
            ret = _original()
            ret.append(self)
            return ret

        # Ditto here
        def wrapped_walk_commands(*, _original=cog.walk_commands):
            yield from _original()
            yield self

        functools.update_wrapper(wrapped_get_commands, cog.get_commands)
        functools.update_wrapper(wrapped_walk_commands, cog.walk_commands)
        cog.get_commands = wrapped_get_commands
        cog.walk_commands = wrapped_walk_commands
        self.cog = cog

    def _eject_cog(self) -> None:
        if self.cog is None:
            return

        # revert back into their original methods
        cog = self.cog
        cog.get_commands = cog.get_commands.__wrapped__
        cog.walk_commands = cog.walk_commands.__wrapped__
        self.cog = None


class HelpCommand:
    r"""The base implementation for help command formatting.

    .. note::

        Internally instances of this class are deep copied every time
        the command itself is invoked to prevent a race condition
        mentioned in discord.py issue 2123.

        This means that relying on the state of this class to be
        the same between command invocations would not work as expected.

    Attributes
    -----------
    context: Optional[:class:`Context`]
        The context that invoked this help formatter. This is generally set
        after the help command assigned, :func:`command_callback`\, has been
        called.
    show_hidden: :class:`bool`
        Specifies if hidden commands should be shown in the output.
        Defaults to ``False``.
    verify_checks: :class:`bool`
        Specifies if commands should have their :attr:`.Command.checks` called
        and verified. Defaults to ``True``.
    command_attrs: :class:`dict`
        A dictionary of options to pass in for the construction of the help
        command. This allows you to change the command behaviour without
        actually changing the implementation of the command. The attributes
        will be the same as the ones passed in the :class:`.Command`
        constructor.
    """

    def __new__(cls, *args: list, **kwargs: dict) -> 'HelpCommand':
        # To prevent race conditions of a single instance while also allowing
        # for settings to be passed the original arguments passed must be
        # assigned to allow for easier copies (which will be made when the
        # help command is actually called)
        # see discord.py issue 2123
        self = super().__new__(cls)

        # Shallow copies cannot be used in this case since it is not unusual
        # to pass instances that need state, e.g. Paginator or what have you
        # into the function. The keys can be safely copied as-is since they're
        # 99.99% certain of being string keys
        deepcopy = copy.deepcopy
        self.__original_kwargs__ = {
            k: deepcopy(v)
            for k, v in kwargs.items()
        }
        self.__original_args__ = deepcopy(args)
        return self

    def __init__(self, **options: dict) -> None:
        self.show_hidden = options.pop('show_hidden', False)
        self.verify_checks = options.pop('verify_checks', True)
        self.command_attrs = attrs = options.pop('command_attrs', {})
        attrs.setdefault('name', 'help')
        attrs.setdefault('help', 'Shows this message')
        self.context = None
        self._command_impl = None

    def copy(self) -> 'HelpCommand':
        o = self.__class__(*self.__original_args__, **self.__original_kwargs__)
        o._command_impl = self._command_impl
        return o

    def _add_to_bot(self, bot: 'Bot') -> None:
        command = _HelpCommandImpl(self, **self.command_attrs)
        bot.add_command(command)
        self._command_impl = command

    def _remove_from_bot(self, bot: 'Bot') -> None:
        bot.remove_command(self._command_impl.name)
        self._command_impl._eject_cog()
        self._command_impl = None

    def get_bot_mapping(self) -> Dict[Optional[Cog], List[Command]]:
        """Retrieves the bot mapping passed to :meth:`send_bot_help`."""
        bot = self.context.bot
        mapping = {
            cog: cog.get_commands()
            for cog in bot.cogs.values()
        }
        mapping[None] = [c for c in bot.all_commands.values() if c.cog is None]
        return mapping

    @property
    def command_prefix(self) -> str:
        """The prefix used to invoke the help command."""
        return self.context.prefix

    @property
    def invoked_with(self) -> str:
        """Similar to :attr:`Context.invoked_with` except properly handles
        the case where :meth:`Context.send_help` is used.

        If the help command was used regularly then this returns
        the :attr:`Context.invoked_with` attribute. Otherwise, if
        it the help command was called using :meth:`Context.send_help`
        then it returns the internal command name of the help command.

        Returns
        ---------
        :class:`str`
            The command name that triggered this invocation.
        """

        command_name = self._command_impl.name
        ctx = self.context
        if (ctx is None or ctx.command is None
                or ctx.command.qualified_name != command_name):
            return command_name
        return ctx.invoked_with

    def get_command_signature(self, command: Command) -> str:
        """Retrieves the signature portion of the help page.

        Parameters
        ----------
        command: :class:`Command`
            The command to get the signature of.

        Returns
        -------
        :class:`str`
            The signature for the command.
        """

        parent = command.full_parent_name
        if len(command.aliases) > 0:
            aliases = '|'.join(command.aliases)
            fmt = '[%s|%s]' % (command.name, aliases)
            if parent:
                fmt = parent + ' ' + fmt
            alias = fmt
        else:
            alias = command.name if not parent else parent + ' ' + command.name

        return '%s%s %s' % (self.command_prefix, alias, command.signature)

    @property
    def cog(self) -> Optional[Cog]:
        """A property for retrieving or setting the cog for the help command.

        When a cog is set for the help command, it is as-if the help command
        belongs to that cog. All cog special methods will apply to the help
        command and it will be automatically unset on unload.

        To unbind the cog from the help command, you can set it to ``None``.

        Returns
        --------
        Optional[:class:`Cog`]
            The cog that is currently set for the help command.
        """
        return self._command_impl.cog

    @cog.setter
    def cog(self, cog: Cog) -> None:
        # Remove whatever cog is currently valid, if any
        self._command_impl._eject_cog()

        # If a new cog is set then inject it.
        if cog is not None:
            self._command_impl._inject_into_cog(cog)

    def command_not_found(self, string: str) -> str:
        """|maybecoro|

        A method called when a command is not found in the help command.
        This is useful to override for i18n.

        Defaults to ``No command called {0} found.``

        Parameters
        ------------
        string: :class:`str`
            The string that contains the invalid command. Note that this has
            had mentions removed to prevent abuse.

        Returns
        ---------
        :class:`str`
            The string to use when a command has not been found.
        """
        return 'No command called "{}" found.'.format(string)

    def subcommand_not_found(self, command: Command, string: str) -> str:
        """|maybecoro|

        A method called when a command did not have a subcommand requested in
        the help command. This is useful to override for i18n.

        Defaults to either:

        - ``'Command "{command.qualified_name}" has no subcommands.'``
            - If there is no subcommand in the ``command`` parameter.
        - ``'Command "{command.qualified_name}" has no subcommand named {string}'``
            - If the ``command`` parameter has subcommands but not one named ``string``.

        Parameters
        ------------
        command: :class:`Command`
            The command that did not have the subcommand requested.
        string: :class:`str`
            The string that contains the invalid subcommand.

        Returns
        ---------
        :class:`str`
            The string to use when the command did not have the subcommand
            requested.
        """  # noqa

        if isinstance(command, Group) and len(command.all_commands) > 0:
            return ('Command "{0.qualified_name}" has no subcommand named '
                    '{1}'.format(command, string))
        return 'Command "{0.qualified_name}" has no subcommands.'.format(
            command
        )

    async def filter_commands(self, commands: Iterable[Command], *,
                              sort: bool = False,
                              key: Optional[Callable] = None
                              ) -> List[Command]:
        """|coro|

        Returns a filtered list of commands and optionally sorts them.

        This takes into account the :attr:`verify_checks` and
        :attr:`show_hidden` attributes.

        Parameters
        ------------
        commands: Iterable[:class:`Command`]
            An iterable of commands that are getting filtered.
        sort: :class:`bool`
            Whether to sort the result.
        key: Optional[Callable[:class:`Command`, Any]]
            An optional key function to pass to :func:`py:sorted` that
            takes a :class:`Command` as its sole parameter. If ``sort`` is
            passed as ``True`` then this will default as the command name.

        Returns
        ---------
        List[:class:`Command`]
            A list of commands that passed the filter.
        """

        if sort and key is None:
            key = lambda c: c.name  # noqa

        if self.show_hidden:
            iterator = commands
        else:
            iterator = filter(lambda c: not c.hidden, commands)

        if not self.verify_checks:
            # if we do not need to verify the checks then we can just
            # run it straight through normally without using await.
            return sorted(iterator, key=key) if sort else list(iterator)

        # if we're here then we need to check every command if it can run
        async def predicate(cmd):
            try:
                return await cmd.can_run(self.context)
            except CommandError:
                return False

        ret = []
        for cmd in iterator:
            valid = await predicate(cmd)
            if valid:
                ret.append(cmd)

        if sort:
            ret.sort(key=key)

        return ret

    def get_max_size(self, commands: Sequence[Command]) -> int:
        """Returns the largest name length of the specified command list.

        Parameters
        ------------
        commands: Sequence[:class:`Command`]
            A sequence of commands to check for the largest size.

        Returns
        --------
        :class:`int`
            The maximum width of the commands.
        """

        as_lengths = (
            _string_width(c.name)
            for c in commands
        )
        return max(as_lengths, default=0)

    def get_destination(self) -> Union[Friend, ClientParty]:
        """Returns either :class:`fortnitepy.Friend` or
        :class:`fortnitepy.ClientParty` where the help command will be output.

        You can override this method to customise the behaviour.

        By default this returns the context's destination.
        """
        return self.context.get_destination()

    async def send_error_message(self, error: Exception) -> None:
        """|coro|

        Handles the implementation when an error happens in the help command.
        For example, the result of :meth:`command_not_found` or
        :meth:`command_has_no_subcommand_found` will be passed here.

        You can override this method to customise the behaviour.

        By default, this sends the error message to the destination
        specified by :meth:`get_destination`.

        .. note::

            You can access the invocation context with
            :attr:`HelpCommand.context`.

        Parameters
        ------------
        error: :class:`str`
            The error message to display to the user.
        """

        destination = self.get_destination()
        await destination.send(error)

    @_not_overridden
    async def help_command_error_handler(self, ctx: Context,
                                         error: Exception) -> None:
        """|coro|

        The help command's error handler, as specified by
        :ref:`ext_commands_error_handler`.

        Useful to override if you need some specific behaviour when the
        error handler is called.

        By default this method does nothing and just propagates to the default
        error handlers.

        Parameters
        ------------
        ctx: :class:`Context`
            The invocation context.
        error: :class:`CommandError`
            The error that was raised.
        """
        pass

    async def send_bot_help(self, page: int) -> None:
        """|coro|

        Handles the implementation of the bot command page in the help command.
        This function is called when the help command is called with no
        arguments.

        It should be noted that this method does not return anything -- rather
        the actual message sending should be done inside this method. Well
        behaved subclasses should use :meth:`get_destination` to know where to
        send, as this is a customisation point for other users.

        You can override this method to customise the behaviour.

        .. note::

            You can access the invocation context with
            :attr:`HelpCommand.context`. Also, the commands in the mapping are
            not filtered. To do the filtering you will have to call
            :meth:`filter_commands` yourself.

        Parameters
        ----------
        page: :class:`int`
            The page to send.
        """
        return None

    async def send_cog_help(self, cog: Cog, page: int) -> None:
        """|coro|

        Handles the implementation of the cog page in the help command.
        This function is called when the help command is called with a cog as
        the argument.

        It should be noted that this method does not return anything -- rather
        the actual message sending should be done inside this method. Well
        behaved subclasses should use :meth:`get_destination` to know where to
        send, as this is a customisation point for other users.

        You can override this method to customise the behaviour.

        .. note::

            You can access the invocation context with
            :attr:`HelpCommand.context`. To get the commands that belong to
            this cog see :meth:`Cog.get_commands`. The commands returned not
            filtered. To do the filtering you will have to call
            :meth:`filter_commands` yourself.

        Parameters
        -----------
        cog: :class:`Cog`
            The cog that was requested for help.
        page: :class:`int`
            The page to send.
        """
        return None

    async def send_group_help(self, group: Group) -> None:
        """|coro|

        Handles the implementation of the group page in the help command.
        This function is called when the help command is called with a group
        as the argument.

        It should be noted that this method does not return anything -- rather
        the actual message sending should be done inside this method. Well
        behaved subclasses should use :meth:`get_destination` to know where to
        send, as this is a customisation point for other users.

        You can override this method to customise the behaviour.

        .. note::

            You can access the invocation context with
            :attr:`HelpCommand.context`. To get the commands that belong to
            this group without aliases see :attr:`Group.commands`. The
            commands returned not filtered. To do the filtering you will have
            to call :meth:`filter_commands` yourself.

        Parameters
        -----------
        group: :class:`Group`
            The group that was requested for help.
        """
        return None

    async def send_command_help(self, command: Command) -> None:
        """|coro|

        Handles the implementation of the single command page in the help
        command.

        It should be noted that this method does not return anything -- rather
        the actual message sending should be done inside this method. Well
        behaved subclasses should use :meth:`get_destination` to know where to
        send, as this is a customisation point for other users.

        You can override this method to customise the behaviour.

        .. note::

            You can access the invocation context with
            :attr:`HelpCommand.context`.

        .. admonition:: Showing Help
            :class: helpful

            There are certain attributes and methods that are helpful for a
            help command to show such as the following:

            - :attr:`Command.help`
            - :attr:`Command.brief`
            - :attr:`Command.short_doc`
            - :attr:`Command.description`
            - :meth:`get_command_signature`

            There are more than just these attributes but feel free to play
            around with these to help you get started to get the output that
            you want.

        Parameters
        -----------
        command: :class:`Command`
            The command that was requested for help.
        """
        return None

    async def prepare_help_command(self, ctx: Context,
                                   command: Optional[Command] = None) -> None:
        """|coro|

        A low level method that can be used to prepare the help command
        before it does anything. For example, if you need to prepare
        some state in your subclass before the command does its processing
        then this would be the place to do it.

        The default implementation does nothing.

        .. note::

            This is called *inside* the help command callback body. So all
            the usual rules that happen inside apply here as well.

        Parameters
        -----------
        ctx: :class:`Context`
            The invocation context.
        command: Optional[:class:`str`]
            The argument passed to the help command.
        """
        pass

    # Not typehinting because its a command callback
    async def command_callback(self, ctx, *, command=None, page: int = 1):
        """|coro|

        The actual implementation of the help command.

        It is not recommended to override this method and instead change
        the behaviour through the methods that actually get dispatched.

        - :meth:`send_bot_help`
        - :meth:`send_cog_help`
        - :meth:`send_group_help`
        - :meth:`send_command_help`
        - :meth:`get_destination`
        - :meth:`command_not_found`
        - :meth:`subcommand_not_found`
        - :meth:`send_error_message`
        - :meth:`on_help_command_error`
        - :meth:`prepare_help_command`
        """

        # page will never get a value but we just include it here for
        # the param list. The actual conversion is done below.
        if command is not None:
            split = command.split()
            try:
                page = int(split[-1])
            except ValueError:
                page = 1
                new = command
            else:
                new = None if len(split) == 1 else ' '.join(split[:-1])
        else:
            new = command

        await self.prepare_help_command(ctx, command)
        bot = ctx.bot

        if new is None:
            # mapping = self.get_bot_mapping()
            return await self.send_bot_help(page)

        # Check if it's a cog
        if not command.startswith(self.command_prefix):
            cog = bot.get_cog(new)
            if cog is not None:
                return await self.send_cog_help(cog, page)

        if command.startswith(self.command_prefix):
            command = command[len(self.command_prefix):]

        maybe_coro = maybe_coroutine

        # If it's not a cog then it's a command.
        # Since we want to have detailed errors when someone
        # passes an invalid subcommand, we need to walk through
        # the command group chain ourselves.
        keys = command.split(' ')
        cmd = bot.all_commands.get(keys[0])
        if cmd is None:
            string = await maybe_coro(self.command_not_found, keys[0])
            return await self.send_error_message(string)

        for key in keys[1:]:
            try:
                found = cmd.all_commands.get(key)
            except AttributeError:
                string = await maybe_coro(self.subcommand_not_found, cmd, key)
                return await self.send_error_message(string)
            else:
                if found is None:
                    string = await maybe_coro(
                        self.subcommand_not_found,
                        cmd,
                        key
                    )
                    return await self.send_error_message(string)
                cmd = found

        if isinstance(cmd, Group):
            return await self.send_group_help(cmd)
        else:
            return await self.send_command_help(cmd)


class FortniteHelpCommand(HelpCommand):
    """The implementation of the default help command.

    This inherits from :class:`HelpCommand`.

    It extends it with the following attributes.

    Attributes
    ------------
    dm_help: Optional[:class:`bool`]
        A tribool that indicates if the help command should DM the user
        instead of sending it to the channel it received it from. If the
        boolean is set to ``True``, then all help output is DM'd. If ``False``,
        none of the help output is DM'd.
    paginator: :class:`Paginator`
        The paginator used to paginate the help command output.
    commands_title: :class:`str`
        The commands title. Defaults to ``Commands:``.
    cog_title: :class:`str`
        The cog title. Defaults to ``Category:``.
    usage_title: :class:`str`
        The usage title. Defaults to ``Usage:``.
    description_title: :class:`str`
        The description title. Defaults to ``Description:``.
    help_title: :class:`str`
        The help title. Defaults to ``Help:``.
    sub_commands_title: :class:`str`
        The sub commands title. Defaults to ``Help Commands:``.
    no_category_heading: :class:`str`
        The text to use as heading if no category (cog) is found
        for the command.
        Defaults to ``No Category``.
    height: :class:`int`
        The maximum number of lines to fit.
        Defaults to ``15``.
    width: :class:`int`
        The maximum number of characters that fit in a line.
        Defaults to ``60``.
    indent: :class:`int`
        How much to indent the commands and other text from a title.
        Defaults to ``4``.
    title_prefix: :class:`str`
        The prefix to use for the help title.
        Defaults to `` +``.
    title_suffix: :class:`str`
        The suffix to use for the help title.
        Defaults to ``+``.
    title_char: :class:`str`
        The char to use for the help title.
        Defaults to ``=``.
    line_prefix: :class:`str`
        The prefix to use for all lines.
        Defaults to ``   ``. (Three spaces)
    line_suffix: :class:`str`
        The prefix to use for all lines.
        Defaults to ````. (Empty)
    footer_prefix: :class:`str`
        The prefix to use for the help footer.
        Defaults to `` +``.
    footer_suffix: :class:`str`
        The suffix to use for the help footer.
        Defaults to ``+``.
    footer_char: :class:`str`
        The char to use for the help footer.
        Defaults to ``=``.
    """

    def __init__(self, **options: dict) -> None:
        self.dm_help = options.pop('dm_help', False)
        self.paginator = options.pop('paginator', None)

        self.commands_title = options.pop('commands_title', 'Commands:')
        self.cog_title = options.pop('cog_title', 'Category:')
        self.usage_title = options.pop('usage_title', 'Usage:')
        self.description_title = options.pop('description_title', 'Description:')  # noqa
        self.help_title = options.pop('help_title', 'Help:')
        self.sub_commands_title = options.pop('sub_commands_title', 'Sub Commands:')  # noqa

        self.no_category = options.pop('no_category_heading', 'No Category')

        self.height = options.pop('height', 15)
        self.width = options.pop('width', 60)
        self.indent = options.pop('indent', 4)

        self.title_prefix = options.pop('title_prefix', ' +')
        self.title_suffix = options.pop('title_suffix', '+')
        self.title_char = options.pop('title_char', '=')

        self.line_prefix = options.pop('line_prefix', '   ')
        self.line_suffix = options.pop('line_suffix', '')

        self.footer_prefix = options.pop('footer_prefix', ' +')
        self.footer_suffix = options.pop('footer_suffix', '+')
        self.footer_char = options.pop('footer_char', '=')

        if self.paginator is None:
            self.paginator = Paginator()

        super().__init__(**options)

    def get_command_name(self, command: Command) -> str:
        """Gets the name of a command.

        This method can be overridden for custom text.

        Parameters
        ----------
        command: :class:`.Command`
            The command to get the name for.

        Returns
        -------
        :class:`str`
            | The command name.
            | Defaults to ``self.command_prefix + command.qualified_name``
        """
        return self.command_prefix + command.qualified_name

    def get_sub_command_name(self, sub_command: Command) -> str:
        """Gets the name of a sub command.

        This method can be overridden for custom text.

        Parameters
        ----------
        sub_command: :class:`.Command`
            The sub command to get the name for.

        Returns
        -------
        :class:`str`
            | The sub command name.
            | Defaults to ``{self.command_prefix} {sub_command.qualified_name}``
        """  # noqa 
        return self.command_prefix + sub_command.qualified_name

    def get_bot_header(self, page_num: int, pages_amount: int) -> str:
        """Gets the name of a sub command.

        This method can be overridden for custom text.

        Parameters
        ----------
        page_num: :class:`int`
            The page being built.
        pages_amount: :class:`int`
            The amount of pages available.

        Returns
        -------
        :class:`str`
            | The sub command name.
            | Defaults to ``{self.command_prefix} {sub_command.qualified_name}``
        """  # noqa

        return '{0} - {1} / {2}'.format(
            'All Commands',
            page_num,
            pages_amount
        )

    def get_bot_footer(self, page_num: int, pages_amount: str) -> str:
        """Gets the text to appear in the footer when
        :meth:`send_bot_help()` is called.

        This method can be overridden for custom text.

        Parameters
        ----------
        page_num: :class:`int`
            The page being built.
        pages_amount: :class:`int`
            The amount of pages available.

        Returns
        -------
        :class:`str`
            | The bot footer.
            | Defaults to ```` (Empty)
        """
        return ''

    def get_command_header(self, command: Command) -> str:
        """Gets the text to appear in the header when
        :meth:`send_command_help()` is called.

        This method can be overridden for custom text.

        Parameters
        ----------
        command: :class:`.Command`
            The command to get the header for.

        Returns
        -------
        :class:`str`
            | The header text.
            | Defaults to ``Command | {self.command_prefix}{command.qualified_name}``
        """  # noqa
        return 'Command | {0}{1}'.format(
            self.command_prefix,
            command.qualified_name
        )

    def get_command_footer(self, command: Command) -> str:
        """Gets the text to appear in the footer when
        :meth:`send_command_help()` is called.

        This method can be overridden for custom text.

        Parameters
        ----------
        command: :class:`.Command`
            The command to get the footer for.

        Returns
        -------
        :class:`str`
            | The footer text.
            | Defaults to ```` (Empty)
        """
        return ''

    def get_group_header(self, group: Group) -> str:
        """Gets the text to appear in the header when
        :meth:`send_group_help()` is called.

        This method can be overridden for custom text.

        Parameters
        ----------
        command: :class:`.Group`
            The group to get the header for.

        Returns
        -------
        :class:`str`
            | The header text.
            | Defaults to ``Command | {self.command_prefix}{group.qualified_name}``
        """  # noqa
        return 'Command | {0}{1}'.format(
            self.command_prefix,
            group.qualified_name
        )

    def get_group_footer(self, group: Group) -> str:
        """Gets the text to appear in the footer when
        :meth:`send_group_help()` is called.

        This method can be overridden for custom text.

        Parameters
        ----------
        command: :class:`.Group`
            The group to get the footer for.

        Returns
        -------
        :class:`str`
            | The footer text.
            | Defaults to ```` (Empty)
        """
        return ''

    def get_cog_header(self, cog: Cog,
                       page_num: int,
                       pages_amount: int) -> str:
        """Gets the text to appear in the header when
        :meth:`send_cog_help()` is called.

        This method can be overridden for custom text.

        Parameters
        ----------
        cog: :class:`.Cog`
            The cog to get the header for.
        page_num: :class:`int`
            The page being built.
        pages_amount: :class:`int`
            The amount of pages available.

        Returns
        -------
        :class:`str`
            | The header text.
            | Defaults to ``Category | {cog.qualified_name} - {page_num} / {pages_amount}``
        """  # noqa
        return 'Category | {0} - {1} / {2}'.format(
            cog.qualified_name,
            page_num,
            pages_amount
        )

    def get_cog_footer(self, cog: Cog,
                       page_num: int,
                       pages_amount: int) -> str:
        """Gets the text to appear in the footer when
        :meth:`send_cog_help()` is called.

        This method can be overridden for custom text.

        Parameters
        ----------
        cog: :class:`.Cog`
            The cog to get the footer for.
        page_num: :class:`int`
            The page being built.
        pages_amount: :class:`int`
            The amount of pages available.

        Returns
        -------
        :class:`str`
            | The footer text.
            | Defaults to ``{self.command_prefix}{self.invoked_with} {cog.qualified_name} <page> | {self.command_prefix}{self.invoked_with} <command>``
        """  # noqa
        return '{0}{1} {2} <page> | {0}{1} <command>'.format(
            self.command_prefix,
            self.invoked_with,
            cog.qualified_name
        )

    def shorten_text(self, text: str,
                     max_len: int,
                     dot_amount: int = 3) -> str:
        """Shortens text to fit into the :attr:`width`."""

        if len(text) > max_len:
            return text[:max_len-dot_amount] + '.'*dot_amount
        return text

    def construct_title(self, t: str) -> str:
        _title = ' ' + t + ' ' if t else ''
        w = self.width - len(self.title_prefix) - len(self.title_suffix)
        return '{0}{1:{2}^{3}}{4}'.format(
            self.title_prefix,
            _title,
            self.title_char,
            w,
            self.title_suffix
        )

    def construct_footer(self, f: str) -> str:
        _footer = ' ' + f + ' ' if f else ''
        w = self.width - len(self.footer_prefix) - len(self.footer_suffix)
        return '{0}{1:{2}^{3}}{4}'.format(
            self.footer_prefix,
            _footer,
            self.footer_char,
            w,
            self.footer_suffix
        )

    def fix_too_long(self, string: str,
                     length: int,
                     start_length: int) -> Tuple[str, List[str]]:
        first = string[:start_length-1]
        string = string[start_length-1:]

        return (
            first,
            [string[0+i:length-1+i] for i in range(0, len(string), length-1)]
        )

    def chunkstring(self, string: str, length: int) -> List[str]:
        lines = []
        curr = ''
        split = string.split()
        for c, word in enumerate(split, 1):
            spaces = 1 if c != len(split) else 0
            if len(word) + spaces > length:
                space_left = (length - len(curr))
                start_length = space_left if space_left > 5 else 0
                first, too_long = self.fix_too_long(word, length, start_length)
                if first:
                    curr += first + '-'

                if curr:
                    lines.append(curr)
                    curr = ''

                for cc, new in enumerate(too_long, 1):
                    if cc != len(too_long):
                        new += '-'
                        lines.append(new)
                    else:
                        curr += new
                continue

            if len(curr) + len(word) > length:
                lines.append(curr[:-1])
                curr = ''

            curr += word + ' '

        if curr:
            lines.append(curr)

        return lines

    def construct_single_line(self, text: str, extra_indent: int = 0) -> str:
        prefix = self.line_prefix + ' '*extra_indent
        suffix = self.line_suffix

        w = self.width - len(prefix) - len(suffix)
        return '{0}{1:<{2}}{3}'.format(
            prefix,
            text,
            w,
            suffix
        )

    def construct_category(self, name: str,
                           brief: str,
                           extra_indent: int = 0,
                           raw: bool = False) -> List[str]:
        prefix = self.line_prefix + ' '*extra_indent
        suffix = self.line_suffix

        indent = self.indent

        w = self.width - len(prefix) - len(suffix)
        name_line = '{0}{1:<{2}}{3}'.format(
            prefix,
            self.shorten_text(name, w),
            w,
            suffix
        )

        brief_w = w - indent
        lines = [name_line]

        if not raw:
            gen = self.chunkstring(brief, brief_w)
        else:
            gen = brief.splitlines()

        for c, line in enumerate(gen, 2):
            fmt = '{0}{1}{2:<{3}}{4}'.format(
                prefix,
                ' '*indent,
                line,
                brief_w,
                suffix
            )
            if c == self.height - 2:
                to_cut = 3 + len(suffix)
                new = fmt[:to_cut] + '...' + suffix
                lines.append(new)
                break

            lines.append(fmt)

        return lines

    async def send_pages(self) -> None:
        """A helper utility to send the page output from :attr:`paginator` to
        the destination.
        """

        destination = self.get_destination()
        for page in self.paginator.pages:
            await destination.send(page)

    async def send_page(self, page_num: int) -> None:
        """A helper utility to send a page output from :attr:`paginator` to
        the destination.
        """

        pages = self.paginator.pages
        if page_num <= 0 or page_num > len(pages):
            return await self.send_error_message(
                'Could not find the page you were looking for'
            )

        destination = self.get_destination()
        await destination.send(pages[page_num-1])

    def get_destination(self) -> Union[Friend, ClientParty]:
        ctx = self.context
        if self.dm_help is True:
            return ctx.author
        elif (self.dm_help is None
                and len(self.paginator) > self.dm_help_threshold):
            return ctx.author
        else:
            return ctx.get_destination()

    async def prepare_help_command(self, ctx: Context,
                                   command: Command) -> None:
        self.paginator.clear()
        await super().prepare_help_command(ctx, command)

    def construct_command_help(self, command: Command) -> List[str]:
        fmt = {}
        if command.cog:
            fmt[self.cog_title] = command.cog.qualified_name

        fmt[self.usage_title] = self.get_command_signature(command)

        if command.description:
            fmt[self.description_title] = command.description

        result = []
        for title, value in fmt.items():
            lines = self.construct_category(title, value)

            result.extend(lines)

        if command.help:
            title = self.help_title
            value = command.help
            lines = self.construct_category(title, value, raw=True)

            result.extend(lines)

        return result

    async def send_bot_help(self, page: int) -> None:
        ctx = self.context
        bot = ctx.bot

        no_category = '\u200b{0.no_category}:'.format(self)

        def get_category(command, *, no_category=no_category):
            cog = command.cog
            return cog.qualified_name if cog is not None else no_category

        filtered = await self.filter_commands(
            bot.commands,
            sort=True,
            key=get_category
        )

        chunks = []
        curr = []

        if bot.description:
            parts = self.construct_category(
                self.description_title,
                bot.description
            )
            curr.extend(parts)

        for command in filtered:
            name = self.get_command_name(command)
            brief = command.brief or ''
            lines = self.construct_category(name, brief)

            if len(lines) + len(curr) > self.height - 2:
                chunks.append(curr)
                curr = []

            curr.extend(lines)

        if curr:
            chunks.append(curr)

        chunks_length = len(chunks)
        for c, chunk in enumerate(chunks, 1):

            footer_fmt = self.get_bot_footer(c, chunks_length) or ''
            page_chunks = [
                self.construct_title(
                    self.get_bot_header(c, chunks_length) or ''
                ),
                *chunk,
                self.construct_footer(footer_fmt.format(
                    self.command_prefix,
                    self.invoked_with,
                ))
            ]
            self.paginator.add_page(
                '\u200b\n' + '\n'.join(page_chunks)
            )

        await self.send_page(page)

    async def send_command_help(self, command: Command) -> None:
        result = self.construct_command_help(command)

        title = self.construct_title(self.get_command_header(command) or '')
        footer = self.construct_footer(self.get_command_footer(command) or '')
        self.paginator.add_page(
            '\u200b\n' + '\n'.join([title, *result, footer])
        )

        await self.send_pages()

    async def send_group_help(self, group: Group) -> None:
        result = self.construct_command_help(group)

        filtered = await self.filter_commands(
            group.commands,
            sort=True
        )

        for c, command in enumerate(filtered):
            if c == 0:
                title = self.sub_commands_title
                result.append('\n'+self.construct_single_line(title))

            name = self.get_sub_command_name(command)
            brief = command.brief or ''
            lines = self.construct_category(
                name,
                brief,
                extra_indent=self.indent
            )

            result.extend(lines)

        title = self.construct_title(
            self.get_group_header(group)
        )
        footer = self.construct_footer('')
        self.paginator.add_page(
            '\u200b\n' + '\n'.join([title, *result, footer])
        )

        await self.send_pages()

    async def send_cog_help(self, cog: Cog, page: str) -> None:
        filtered = await self.filter_commands(
            cog.get_commands(),
            sort=True
        )

        chunks = []
        curr = []

        if cog.description:
            parts = self.construct_category(
                self.description_title,
                cog.description
            )
            curr.extend(parts)

        for c, command in enumerate(filtered):
            if c == 0:
                title = self.commands_title
                pre = '\n' if curr else ''
                curr.append(pre+self.construct_single_line(title))

            name = self.get_command_name(command)
            brief = command.brief or ''
            lines = self.construct_category(
                name,
                brief,
                extra_indent=self.indent
            )

            if len(lines) + len(curr) > self.height - 2:
                chunks.append(curr)
                curr = []

            curr.extend(lines)

        if curr:
            chunks.append(curr)

        chunks_length = len(chunks)
        for c, chunk in enumerate(chunks, 1):
            title = self.construct_title(
                self.get_cog_header(cog, c, chunks_length) or ''
            )
            fmt = self.get_cog_footer(cog, c, chunks_length) or ''
            footer = self.construct_footer(fmt)
            page_chunks = [
                title,
                *chunk,
                footer
            ]
            self.paginator.add_page(
                '\u200b\n' + '\n'.join(page_chunks)
            )

        await self.send_page(page)



# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019-2021 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json
import re
import datetime

from typing import TYPE_CHECKING

from .errors import Forbidden, PartyError
from .enums import Platform
from .kairos import Avatar

if TYPE_CHECKING:
    from .client import Client
    from .friend import Friend
    from .party import ClientParty


class PresenceGameplayStats:
    """Represents gameplaystats received from presence.

    Attributes
    ----------
    friend: :class:`Friend`
        The friend these stats belong to.
    state: :class:`str`
        The state.

        .. note::

            It's not really known what value this property might
            hold. This is pretty much always an empty string.
    playlist: :class:`str`
        The playlist.

        .. note::

            The playlist from the gameplay stats property usually
            isn't updated. Consider using :attr:`Presence.playlist` instead
            as that seems to always be the correct playlist.
    players_alive: :class:`int`
        The amount of players alive in the current game.
    kills: :class:`int`
        The amount of kills the friend currently has. Aliased to ``num_kills``
        as well for legacy reasons.
    fell_to_death: :class:`bool`
        ``True`` if friend fell to death in its current game, else ``False``
    """

    __slots__ = ('friend', 'state', 'playlist', 'players_alive', 'kills',
                 'num_kills', 'fell_to_death')

    def __init__(self, friend: 'Friend',
                 data: str,
                 players_alive: int) -> None:
        self.friend = friend
        self.state = data.get('state')
        self.playlist = data.get('playlist')
        self.players_alive = players_alive

        self.kills = data.get('numKills')
        if self.kills is not None:
            self.kills = int(self.kills)

        self.num_kills = self.kills

        self.fell_to_death = True if data.get('bFellToDeath') else False

    def __repr__(self) -> str:
        return ('<PresenceGameplayStats friend={0.friend!r} '
                'players_alive={0.players_alive} num_kills={0.num_kills} '
                'playlist={0.playlist!r}>'.format(self))


class PresenceParty:
    """Represents a party received from presence.

    Before accessing any of this class' attributes or functions
    you should always check if the party is private: ::

        @client.event
        async def event_friend_presence(before, after):
            # after is the newly received presence
            presence = after

            # check if presence is from the account 'Terbau'
            # NOTE: you should always use id over display_name
            # but for this example i've use display_name just
            # to demonstrate.
            if presence.friend.display_name != 'Terbau':
                return

            # check if party is private
            if presence.party.private:
                return

            # if all the checks above succeeds we join the party
            await presence.party.join()


    .. note::

        If the party is private, all attributes below private will
        be ``None``.

    Attributes
    ----------
    client: :class:`str`
        The client.
    private: :class:`bool`
        ``True`` if the party is private else ``False``.
    platform: :class:`Platform`
        The platform of the friend.
    id: :class:`str`
        The party's id.
    party_type_id: :class:`str`
        The party's type id.
    key: :class:`str`
        The party's key.
    app_id: :class:`str`
        The party's app id.
    build_id: :class:`str`
        The party's build id. Similar format to :attr:`Client.party_build_id`.
    net_cl: :class:`str`
        The party's net_cl. Similar format to :attr:`Client.net_cl`.
    party_flags: :class:`str`
        The party's flags.
    not_accepting_reason: :class:`str`
        The party's not accepting reason.
    playercount: :class:`int`
        The party's playercount.
    """

    __slots__ = ('client', 'private', 'platform', 'id', 'party_type_id',
                 'key', 'app_id', 'build_id', 'net_cl', 'party_flags',
                 'not_accepting_reason', 'playercount')

    def __init__(self, client: 'Client', data: dict) -> None:
        self.client = client
        self.private = data.get('bIsPrivate', False)

        pl = data.get('sourcePlatform')
        self.platform = Platform(pl) if pl is not None else None
        self.id = data.get('partyId')
        self.party_type_id = data.get('partyTypeId')
        self.key = data.get('key')
        self.app_id = data.get('appId')
        self.build_id = data.get('buildId')

        if self.build_id is not None and self.build_id.startswith('1:1:'):
            self.net_cl = self.build_id[4:]
        else:
            self.net_cl = None

        self.party_flags = data.get('partyFlags')
        self.not_accepting_reason = data.get('notAcceptingReason')

        self.playercount = data.get('pc')
        if self.playercount is not None:
            self.playercount = int(self.playercount)

    def __repr__(self) -> str:
        return ('<PresenceParty private={0.private} id={0.id!r} '
                'playercount={0.playercount}>'.format(self))

    async def join(self) -> 'ClientParty':
        """|coro|

        Joins the friends' party.

        Raises
        ------
        PartyError
            You are already a member of this party.
        Forbidden
            The party is private.
        HTTPException
            Something else went wrong when trying to join this party.

        Returns
        -------
        :class:`ClientParty`
            The party that was just joined.
        """
        if self.client.party.id == self.id:
            raise PartyError('You are already a member of this party.')

        if self.private:
            raise Forbidden('You cannot join a private party.')

        return await self.client.join_party(self.id)


class Presence:
    """Represents a presence received from a friend

    Attributes
    ----------
    client: :class:`Client`
        The client.
    available: :class:`bool`
        The availability of this presence. ``True`` if presence is available,
        ``False`` if user went unavailable.
    away: :class:`AwayStatus`
        The users away status.
    friend: :class:`Friend`
        The friend you received this presence from.
    platform: :class:`Platform`
        The platform this presence was sent from.
    received_at: :class:`datetime.datetime`
        The UTC time of when the client received this presence.
    status: :class:`str`
        The friend's status.
    in_kairos: :class:`bool`
        Wether or not the friend is in kairos. If this is True, then quite a
        lot of the property attributes will be None or potentially something
        unexpected.
    playing: :class:`bool`
        Says if friend is playing.
    joinable: :class:`bool`
        Says if friend is joinable.
    session_id: :class:`str`
        The friend's current session id. Often referred to as
        server key or game key. Returns ``None`` if the friend is not currently
        in a game.
    has_properties: :class:`bool`
        ``True`` if the presence has properties else ``False``.

        .. warning::

            All attributes below this point will be ``None`` if
            :attr:`has_properties` is ``False``. The only exception is some
            attributes like :attr:`Presence.avatar` and
            :attr:`Presence.party` which isn't None as long as
            :attr:`Presence.in_kairos` is True.
    party: :class:`PresenceParty`
        The friend's party.
    gameplay_stats: Optional[:class:`PresenceGameplayStats`]
        The friend's gameplay stats. Will be ``None`` if no gameplay stats
        are currently availble.
    avatar: :class:`Avatar`
        The avatar set in Kairos (Mobile app).
    homebase_rating: :class:`str`
        The friend's homebase rating
    lfg: :class:`bool`
        ``True`` if the friend is currently looking for a game.
    sub_game: :class:`str`
        The friend's current subgame.
    in_unjoinable_match: :class:`bool`
        ``True`` if friend is in unjoinable match else ``False``.
    playlist: :class:`str`
        The friend's current playlist.
    party_size: :class:`int`
        The size of the friend's party.
    max_party_size: :class:`int`
        The max size of the friend's party.
    game_session_join_key: :class:`str`
        The join key of the friend's session.
    server_player_count: :class:`str`
        The playercount of the friend's server.
    """

    __slots__ = ('client', 'available', 'away', 'friend', 'platform',
                 'received_at', 'status', 'in_kairos', 'playing', 'joinable',
                 'has_voice_support', 'session_id',
                 'has_properties', 'avatar', 'homebase_rating', 'lfg',
                 'sub_game', 'in_unjoinable_match', 'playlist', 'party_size',
                 'max_party_size', 'game_session_join_key',
                 'server_player_count', 'gameplay_stats', 'party')

    def __init__(self, client: 'Client',
                 from_id: str,
                 platform: str,
                 available: bool,
                 away: bool,
                 data: dict) -> None:
        self.client = client
        self.available = available
        self.away = away
        self.friend = self.client.get_friend(from_id)
        self.platform = Platform(platform)
        self.received_at = datetime.datetime.utcnow()

        self.status = data['Status']
        self.in_kairos = data.get('bIsEmbedded', False)
        self.playing = data['bIsPlaying']
        self.joinable = data['bIsJoinable']
        self.has_voice_support = data['bHasVoiceSupport']
        self.session_id = (data['SessionId'] if
                           data['SessionId'] != "" else None)

        raw_properties = data.get('Properties', {})
        self.has_properties = raw_properties != {}

        # All values below will be "None" if properties is empty.
        # The only expections are avatar and party which could have
        # values as long as in_kairos is True.

        kairos_p = raw_properties.get('KairosProfile_s', {})
        if kairos_p:
            kairos_p = json.loads(kairos_p)
        else:
            kairos_p = raw_properties.get('KairosProfile_j', {})

        background = kairos_p.get('avatarBackground')
        if background and not isinstance(background, list):
            background = json.loads(background)

        self.avatar = Avatar(
            asset=kairos_p.get('avatar'),
            background_colors=background
        )

        _basic_info = raw_properties.get('FortBasicInfo_j', {})
        self.homebase_rating = _basic_info.get('homeBaseRating')

        if raw_properties.get('FortLFG_I') is None:
            self.lfg = None
        else:
            self.lfg = int(raw_properties.get('FortLFG_I')) == 1

        self.sub_game = raw_properties.get('FortSubGame_i')

        self.in_unjoinable_match = raw_properties.get(
            'InUnjoinableMatch_b'
        )
        if self.in_unjoinable_match is not None:
            self.in_unjoinable_match = int(self.in_unjoinable_match)

        self.playlist = raw_properties.get('GamePlaylistName_s')

        players_alive = raw_properties.get('Event_PlayersAlive_s')
        if players_alive is not None:
            players_alive = int(players_alive)

        self.party_size = raw_properties.get('Event_PartySize_s')
        if self.party_size is not None:
            self.party_size = int(self.party_size)

        self.max_party_size = raw_properties.get('Event_PartyMaxSize_s')
        if self.max_party_size is not None:
            self.max_party_size = int(self.max_party_size)

        self.game_session_join_key = raw_properties.get(
            'GameSessionJoinKey_s'
        )

        self.server_player_count = raw_properties.get(
            'ServerPlayerCount_i'
        )
        if self.server_player_count is not None:
            self.server_player_count = int(self.server_player_count)

        if 'FortGameplayStats_j' in raw_properties.keys():
            self.gameplay_stats = PresenceGameplayStats(
                self.friend,
                raw_properties['FortGameplayStats_j'],
                players_alive
            )
        else:
            self.gameplay_stats = None

        key = None
        for k in raw_properties.keys():
            if re.search(r'party\.joininfodata\.\d+_j', k) is not None:
                key = k

        if key is None:
            self.party = None
        else:
            self.party = PresenceParty(self.client, raw_properties[key])

    def __repr__(self) -> str:
        return ('<Presence friend={0.friend!r} available={0.available} '
                'received_at={0.received_at!r}>'.format(self))



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


import asyncio
import functools
import inspect
import datetime

from collections import OrderedDict
from typing import Iterable, Optional, Union, Awaitable, Any, Set
from fortnitepy.errors import FortniteException
from fortnitepy.typedefs import MaybeCoro

from . import converter as converters
from . import errors
from ._types import _BaseCommand
from .cog import Cog
from .cooldown import Cooldown, BucketType, CooldownMapping, MaxConcurrency
from .context import Context
from .converter import Converter


__all__ = (
    'Command',
    'Group',
    'GroupMixin',
    'command',
    'group',
    'check',
    'check_any',
    'before_invoke',
    'after_invoke',
    'cooldown',
    'max_concurrency',
    'dm_only',
    'party_only',
    'is_owner',
)


def wrap_callback(coro: Awaitable) -> Awaitable:
    @functools.wraps(coro)
    async def wrapped(*args: list, **kwargs: dict) -> Any:
        try:
            ret = await coro(*args, **kwargs)
        except errors.CommandError:
            raise
        except asyncio.CancelledError:
            return
        except Exception as exc:
            raise errors.CommandInvokeError(exc) from exc
        return ret
    return wrapped


def hooked_wrapped_callback(command: 'Command',
                            ctx: Context,
                            coro: Awaitable) -> Awaitable:
    @functools.wraps(coro)
    async def wrapped(*args: list, **kwargs: dict) -> Awaitable:
        try:
            ret = await coro(*args, **kwargs)
        except errors.CommandError:
            ctx.command_failed = True
            raise
        except asyncio.CancelledError:
            ctx.command_failed = True
            return
        except Exception as exc:
            ctx.command_failed = True
            raise errors.CommandInvokeError(exc) from exc
        finally:
            if command._max_concurrency is not None:
                await command._max_concurrency.release(ctx)

            await command.call_after_hooks(ctx)

        return ret
    return wrapped


def _convert_to_bool(argument: str) -> bool:
    lowered = argument.lower()
    if lowered in ('yes', 'y', 'true', 't', '1', 'enable', 'on'):
        return True
    elif lowered in ('no', 'n', 'false', 'f', '0', 'disable', 'off'):
        return False
    else:
        raise errors.BadArgument(
            lowered + ' is not a recognised boolean option'
        )


class _CaseInsensitiveDict(dict):
    def __contains__(self, k):
        return super().__contains__(k.casefold())

    def __delitem__(self, k):
        return super().__delitem__(k.casefold())

    def __getitem__(self, k):
        return super().__getitem__(k.casefold())

    def get(self, k, default=None):
        return super().get(k.casefold(), default)

    def pop(self, k, default=None):
        return super().pop(k.casefold(), default)

    def __setitem__(self, k, v):
        super().__setitem__(k.casefold(), v)


class Command(_BaseCommand):
    r"""A class that implements the protocol for a bot text command.
    These are not created manually, instead they are created via the
    decorator or functional interface.

    Attributes
    -----------
    name: :class:`str`
        The name of the command.
    callback:
        The coroutine that is executed when the command is called.
    help: :class:`str`
        The long help text for the command.
    brief: :class:`str`
        The short help text for the command. If this is not specified
        then the first line of the long help text is used instead.
    usage: :class:`str`
        A replacement for arguments in the default help text.
    aliases: Union[:class:`list`, :class:`tuple`]
        The list of aliases the command can be invoked under.
    enabled: :class:`bool`
        A boolean that indicates if the command is currently enabled.
        If the command is invoked while it is disabled, then
        :exc:`.DisabledCommand` is raised to the :func:`.event_command_error`
        event. Defaults to ``True``.
    parent: Optional[:class:`Command`]
        The parent command that this command belongs to. ``None`` if there
        isn't one.
    cog: Optional[:class:`Cog`]
        The cog that this command belongs to. ``None`` if there isn't one.
    checks: List[Callable[..., :class:`bool`]]
        A list of predicates that verifies if the command could be executed
        with the given :class:`.Context` as the sole parameter. If an exception
        is necessary to be thrown to signal failure, then one inherited from
        :exc:`.CommandError` should be used. Note that if the checks fail then
        :exc:`.CheckFailure` exception is raised to the
        :func:`.event_command_error` event.
    description: :class:`str`
        The message prefixed into the default help command.
    hidden: :class:`bool`
        If ``True``\, the default help command does not show this in the
        help output.
    rest_is_raw: :class:`bool`
        If ``False`` and a keyword-only argument is provided then the keyword
        only argument is stripped and handled as if it was a regular argument
        that handles :exc:`.MissingRequiredArgument` and default values in a
        regular matter rather than passing the rest completely raw. If ``True``
        then the keyword-only argument will pass in the rest of the arguments
        in a completely raw matter. Defaults to ``False``.
    invoked_subcommand: Optional[:class:`Command`]
        The subcommand that was invoked, if any.
    ignore_extra: :class:`bool`
        If ``True``\, ignores extraneous strings passed to a command if all its
        requirements are met (e.g. ``?foo a b c`` when only expecting ``a``
        and ``b``). Otherwise :func:`.event_command_error` and local error
        handlers are called with :exc:`.TooManyArguments`. Defaults to
        ``True``.
    cooldown_after_parsing: :class:`bool`
        If ``True``\, cooldown processing is done after argument parsing,
        which calls converters. If ``False`` then cooldown processing is done
        first and then the converters are called second. Defaults to ``False``.
    """

    def __new__(cls, *args: list, **kwargs: dict) -> 'Command':
        self = super().__new__(cls)

        self.__fnpy_original_kwargs__ = kwargs.copy()
        return self

    def __init__(self, coro: Awaitable, **kwargs: dict) -> None:
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('Command callback must be a coroutine')

        self.name = kwargs.get('name') or coro.__name__
        if not isinstance(self.name, str):
            raise TypeError('The name of the command must be a string')

        self.callback = coro
        self.enabled = kwargs.get('enabled', True)

        help_doc = kwargs.get('help')
        if help_doc is not None:
            help_doc = inspect.cleandoc(help_doc)
        else:
            help_doc = inspect.getdoc(coro)
            if isinstance(help_doc, bytes):
                help_doc = help_doc.decode('utf-8')

        self.help = help_doc

        self.brief = kwargs.get('brief')
        self.usage = kwargs.get('usage')
        self.rest_is_raw = kwargs.get('rest_is_raw', False)
        self.aliases = kwargs.get('aliases', [])

        if not isinstance(self.aliases, (list, tuple)):
            raise TypeError('aliases must be a list or tuple')

        self.description = inspect.cleandoc(kwargs.get('description', ''))
        self.hidden = kwargs.get('hidden', False)

        try:
            checks = coro.__fnpy_commands_checks__
            checks.reverse()
        except AttributeError:
            checks = kwargs.get('checks', [])
        finally:
            self.checks = checks

        try:
            cooldown = coro.__fnpy_commands_cooldown__
        except AttributeError:
            cooldown = kwargs.get('cooldown')
        finally:
            self._buckets = CooldownMapping(cooldown)

        try:
            max_concurrency = coro.__fnpy_commands_max_concurrency__
        except AttributeError:
            max_concurrency = kwargs.get('max_concurrency')
        finally:
            self._max_concurrency = max_concurrency

        self.ignore_extra = kwargs.get('ignore_extra', True)
        self.cooldown_after_parsing = kwargs.get('cooldown_after_parsing',
                                                 False)
        self.cog = None
        self.instance = None

        # bandaid for the fact that sometimes parent can be the bot instance
        parent = kwargs.get('parent')
        self.parent = parent if isinstance(parent, Command) else None

        try:
            before_invoke = coro.__fnpy_before_invoke__
        except AttributeError:
            self._before_invoke = None
        else:
            self.before_invoke(before_invoke)

        try:
            after_invoke = coro.__fnpy_after_invoke__
        except AttributeError:
            self._after_invoke = None
        else:
            self.after_invoke(after_invoke)

    @property
    def callback(self) -> Awaitable:
        return self._callback

    @callback.setter
    def callback(self, function: Awaitable) -> None:
        self._callback = function
        self.module = function.__module__

        signature = inspect.signature(function)
        self.params = signature.parameters.copy()

        for key, value in self.params.items():
            if isinstance(value.annotation, str):
                self.params[key] = value = value.replace(
                    annotation=eval(value.annotation, function.__globals__)
                )

    @property
    def self_instance(self) -> Union[Cog, object]:
        return self.cog or self.instance

    def add_check(self, func: MaybeCoro) -> None:
        """Adds a check to the command.

        This is the non-decorator interface to :func:`.check`.


        Parameters
        ----------
        func
            The function that will be used as a check.
        """
        self.checks.append(func)

    def remove_check(self, func: MaybeCoro) -> None:
        """Removes a check from the command.

        This function is idempotent and will not raise an exception
        if the function is not in the command's checks.

        Parameters
        -----------
        func
            The function to remove from the checks.
        """

        try:
            self.checks.remove(func)
        except ValueError:
            pass

    def update(self, **kwargs: dict) -> None:
        """Updates :class:`Command` instance with updated attribute.

        This works similarly to the :func:`.command` decorator in terms
        of parameters in that they are passed to the :class:`Command` or
        subclass constructors, sans the name and callback.
        """
        self.__init__(
            self.callback,
            **dict(self.__fnpy_original_kwargs__, **kwargs)
        )

    async def __call__(self, *args: list, **kwargs: dict) -> Any:
        """|coro|

        Calls the internal callback that the command holds.

        .. note::

            This bypasses all mechanisms -- including checks, converters,
            invoke hooks, cooldowns, etc. You must take care to pass
            the proper arguments and types to this function.
        """

        if self.self_instance is not None:
            return await self.callback(self.self_instance, *args, **kwargs)
        else:
            return await self.callback(*args, **kwargs)

    def _ensure_assignment_on_copy(self, other: 'Command') -> 'Command':
        other._before_invoke = self._before_invoke
        other._after_invoke = self._after_invoke
        if self.checks != other.checks:
            other.checks = self.checks.copy()
        if self._buckets.valid and not other._buckets.valid:
            other._buckets = self._buckets.copy()
        if self._max_concurrency != other._max_concurrency:
            other._max_concurrency = self._max_concurrency.copy()

        try:
            other.error_handler = self.error_handler
        except AttributeError:
            pass
        return other

    def copy(self) -> 'Command':
        ret = self.__class__(self.callback, **self.__fnpy_original_kwargs__)
        return self._ensure_assignment_on_copy(ret)

    def _update_copy(self, kwargs: dict) -> 'Command':
        if kwargs:
            kw = kwargs.copy()
            kw.update(self.__fnpy_original_kwargs__)
            copy = self.__class__(self.callback, **kw)
            return self._ensure_assignment_on_copy(copy)
        else:
            return self.copy()

    async def dispatch_error(self, ctx: Context, error: Exception) -> Any:
        ctx.command_failed = True
        cog = self.cog
        instance = self.self_instance

        ret = False

        try:
            coro = self.error_handler
        except AttributeError:
            pass
        else:
            injected = wrap_callback(coro)
            if instance is not None:
                ret = await injected(instance, ctx, error)
            else:
                ret = await injected(ctx, error)

        if ret is False:
            if cog is not None:
                local = Cog._get_overridden_method(cog.cog_command_error)
                if local is not None:
                    wrapped = wrap_callback(local)
                    ret = await wrapped(ctx, error)

        if ret is False:
            ctx.bot.dispatch_error(ctx, error)

    async def _actual_conversion(self,
                                 ctx: Context,
                                 converter: Converter,
                                 argument: str,
                                 param: inspect.Parameter) -> Any:
        if converter is bool:
            return _convert_to_bool(argument)

        try:
            module = converter.__module__
        except AttributeError:
            pass
        else:
            if module is not None and (module.startswith('fortnitepy.')
                                       and not module.endswith('converter')):
                converter = getattr(
                    converters,
                    converter.__name__ + 'Converter',
                    converter
                )

        try:
            if inspect.isclass(converter):
                if issubclass(converter, converters.Converter):
                    instance = converter()
                    ret = await instance.convert(ctx, argument)
                    return ret
                else:
                    method = getattr(converter, 'convert', None)
                    if method is not None and inspect.ismethod(method):
                        ret = await method(ctx, argument)
                        return ret

            elif isinstance(converter, converters.Converter):
                ret = await converter.convert(ctx, argument)
                return ret

        except errors.CommandError:
            raise
        except Exception as exc:
            raise errors.ConversionError(converter, exc) from exc

        try:
            return converter(argument)
        except errors.CommandError:
            raise
        except Exception as exc:
            try:
                name = converter.__name__
            except AttributeError:
                name = converter.__class__.__name__

            raise errors.BadArgument('Converting to "{}" failed for parameter '
                                     '"{}".'.format(name, param.name)) from exc

    async def do_conversion(self, ctx: Context,
                            converter: Converter,
                            argument: str,
                            param: inspect.Parameter) -> Any:
        try:
            origin = converter.__origin__
        except AttributeError:
            pass
        else:
            if origin is Union:
                errors = []
                NoneType = type(None)
                for conv in converter.__args__:
                    if conv is NoneType and param.kind != param.VAR_POSITIONAL:
                        ctx.view.undo()
                        if param.default is param.empty:
                            return None
                        else:
                            return param.default

                    try:
                        value = await self._actual_conversion(
                            ctx,
                            conv,
                            argument,
                            param
                        )
                    except errors.CommandError as exc:
                        errors.append(exc)
                    else:
                        return value

                raise errors.BadUnionArgument(
                    param,
                    converter.__args__,
                    errors
                )

        return await self._actual_conversion(ctx, converter, argument, param)

    def _get_converter(self, param: inspect.Parameter) -> inspect.Parameter:
        conv = param.annotation
        if conv is param.empty:
            if param.default is not param.empty:
                conv = str if param.default is None else type(param.default)
            else:
                conv = str

        return conv

    async def transform(self, ctx: Context, param: inspect.Parameter) -> Any:
        required = param.default is param.empty
        converter = self._get_converter(param)
        consume_rest_is_special = (param.kind == param.KEYWORD_ONLY
                                   and not self.rest_is_raw)
        view = ctx.view
        view.skip_ws()

        if type(converter) is converters._Greedy:
            if param.kind == param.POSITIONAL_OR_KEYWORD:
                return await self._transform_greedy_pos(
                    ctx,
                    param,
                    required,
                    converter.converter
                )
            elif param.kind == param.VAR_POSITIONAL:
                return await self._transform_greedy_var_pos(
                    ctx,
                    param,
                    converter.converter
                )
            else:
                converter = converter.converter

        if view.eof:
            if param.kind == param.VAR_POSITIONAL:
                raise RuntimeError()

            if required:
                if self._is_typing_optional(param.annotation):
                    return None
                raise errors.MissingRequiredArgument(param)
            return param.default

        previous = view.index
        if consume_rest_is_special:
            argument = view.read_rest().strip()
        else:
            argument = view.get_quoted_word()
        view.previous = previous

        return await self.do_conversion(ctx, converter, argument, param)

    async def _transform_greedy_pos(self, ctx: Context,
                                    param: inspect.Parameter,
                                    required: bool,
                                    converter: Converter) -> Any:
        view = ctx.view
        result = []
        while not view.eof:
            previous = view.index

            view.skip_ws()
            try:
                argument = view.get_quoted_word()
                value = await self.do_conversion(
                    ctx,
                    converter,
                    argument,
                    param
                )
            except (errors.CommandError, errors.ArgumentParsingError):
                view.index = previous
                break
            else:
                result.append(value)

        if not result and not required:
            return param.default
        return result

    async def _transform_greedy_var_pos(self, ctx: Context,
                                        param: inspect.Parameter,
                                        converter: Converter) -> Any:
        view = ctx.view
        previous = view.index
        try:
            argument = view.get_quoted_word()
            value = await self.do_conversion(ctx, converter, argument, param)
        except (errors.CommandError, errors.ArgumentParsingError):
            view.index = previous
            raise RuntimeError() from None
        else:
            return value

    @property
    def clean_params(self) -> OrderedDict:
        """OrderedDict[:class:`str`, :class:`typing.Parameter`]: Retrieves the
        parameter OrderedDict without the context or self parameters.

        Useful for inspecting signature.
        """

        result = self.params.copy()
        if self.self_instance is not None:
            result.popitem(last=False)

        try:
            result.popitem(last=False)
        except Exception:
            raise ValueError('Missing context parameter') from None

        return result

    @property
    def full_parent_name(self) -> str:
        """:class:`str`: Retrieves the fully qualified parent command name.

        This is the base command name required to execute it. For example,
        in ``?one two three`` the parent name would be ``one two``.
        """

        elems = []
        command = self
        while command.parent is not None:
            command = command.parent
            elems.append(command.name)

        return ' '.join(reversed(elems))

    @property
    def parents(self) -> 'Command':
        """:class:`Command`: Retrieves the parents of this command.

        If the command has no parents then it returns an empty :class:`list`.

        For example in commands ``?a b c test``, the parents are ``[c, b, a]``.
        """

        elems = []
        command = self
        while command.parent is not None:
            command = command.parent
            elems.append(command)

        return elems

    @property
    def root_parent(self) -> 'Command':
        """Retrieves the root parent of this command.

        If the command has no parents then it returns ``None``.

        For example in commands ``?a b c test``, the root parent is ``a``.
        """

        if not self.parent:
            return None

        return self.parents[-1]

    @property
    def qualified_name(self) -> str:
        """:class:`str`: Retrieves the fully qualified command name.

        This is the full parent name with the command name as well.

        For example, in ``?one two three`` the qualified name would be
        ``one two three``.
        """

        parent_name = self.full_parent_name
        if parent_name:
            return '{} {}'.format(parent_name, self.name)
        else:
            return self.name

    def __str__(self) -> str:
        return self.qualified_name

    async def _parse_arguments(self, ctx: Context) -> None:
        instance = self.self_instance
        ctx.args = [ctx] if instance is None else [instance, ctx]
        ctx.kwargs = {}
        args = ctx.args
        kwargs = ctx.kwargs

        view = ctx.view
        iterator = iter(self.params.items())

        if instance is not None:
            try:
                next(iterator)
            except StopIteration:
                raise FortniteException(
                    'Callback for {0.name} command is missing '
                    '"self" parameter.'.format(self)
                )

        try:
            next(iterator)
        except StopIteration:
            raise FortniteException(
                'Callback for {0.name} command is missing '
                '"ctx" parameter.'.format(self)
            )

        for name, param in iterator:
            if param.kind == param.POSITIONAL_OR_KEYWORD:
                transformed = await self.transform(ctx, param)
                args.append(transformed)

            elif param.kind == param.KEYWORD_ONLY:
                if self.rest_is_raw:
                    converter = self._get_converter(param)
                    argument = view.read_rest()
                    kwargs[name] = await self.do_conversion(
                        ctx,
                        converter,
                        argument,
                        param
                    )
                else:
                    kwargs[name] = await self.transform(ctx, param)
                break

            elif param.kind == param.VAR_POSITIONAL:
                while not view.eof:
                    try:
                        transformed = await self.transform(ctx, param)
                        args.append(transformed)
                    except RuntimeError:
                        break

        if not self.ignore_extra:
            if not view.eof:
                fmt = 'Too many arguments passed to ' + self.qualified_name
                raise errors.TooManyArguments(fmt)

    async def call_before_hooks(self, ctx: Context) -> None:
        inst = self.self_instance
        cog = self.cog
        if self._before_invoke is not None:
            try:
                instance = self._before_invoke.__self__
            except AttributeError:
                if inst:
                    await self._before_invoke(inst, ctx)
                else:
                    await self._before_invoke(ctx)
            else:
                await self._before_invoke(instance, ctx)

        if cog is not None:
            hook = Cog._get_overridden_method(cog.cog_before_invoke)
            if hook is not None:
                await hook(ctx)

        hook = ctx.bot._before_invoke
        if hook is not None:
            await hook(ctx)

    async def call_after_hooks(self, ctx: Context) -> None:
        inst = self.self_instance
        cog = self.cog
        if self._after_invoke is not None:
            try:
                instance = self._after_invoke.__self__
            except AttributeError:
                if inst:
                    await self._after_invoke(inst, ctx)
                else:
                    await self._after_invoke(ctx)
            else:
                await self._after_invoke(instance, ctx)

        if cog is not None:
            hook = Cog._get_overridden_method(cog.cog_after_invoke)
            if hook is not None:
                await hook(ctx)

        hook = ctx.bot._after_invoke
        if hook is not None:
            await hook(ctx)

    def _prepare_cooldowns(self, ctx: Context) -> None:
        if self._buckets.valid:
            current = ctx.message.created_at.replace(
                tzinfo=datetime.timezone.utc
            )
            timestamp = current.timestamp()

            bucket = self._buckets.get_bucket(ctx.message, timestamp)
            retry_after = bucket.update_rate_limit(timestamp)
            if retry_after:
                raise errors.CommandOnCooldown(bucket, retry_after)

    async def prepare(self, ctx: Context) -> None:
        ctx.command = self

        if not await self.can_run(ctx):
            raise errors.CheckFailure(
                'The checks for command {0.qualified_name} failed'.format(self)
            )

        if self.cooldown_after_parsing:
            await self._parse_arguments(ctx)
            self._prepare_cooldowns(ctx)
        else:
            self._prepare_cooldowns(ctx)
            await self._parse_arguments(ctx)

        if self._max_concurrency is not None:
            await self._max_concurrency.acquire(ctx)

        await self.call_before_hooks(ctx)

    def is_on_cooldown(self, ctx: Context) -> bool:
        """Checks whether the command is currently on cooldown.

        Parameters
        ----------
        ctx: :class:`.Context`
            The invocation context to use when checking the commands
            cooldown status.

        Returns
        --------
        :class:`bool`
            A boolean indicating if the command is on cooldown.
        """

        if not self._buckets.valid:
            return False

        bucket = self._buckets.get_bucket(ctx.message)
        return bucket.get_tokens() == 0

    def reset_cooldown(self, ctx: Context) -> None:
        """Resets the cooldown on this command.

        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context to reset the cooldown under.
        """

        if self._buckets.valid:
            bucket = self._buckets.get_bucket(ctx.message)
            bucket.reset()

    async def invoke(self, ctx: Context) -> None:
        await self.prepare(ctx)

        ctx.invoked_subcommand = None
        injected = hooked_wrapped_callback(self, ctx, self.callback)
        await injected(*ctx.args, **ctx.kwargs)

    async def reinvoke(self, ctx: Context, *,
                       call_hooks: bool = False) -> None:
        ctx.command = self
        await self._parse_arguments(ctx)

        if call_hooks:
            await self.call_before_hooks()

        ctx.invoked_subcommand = None
        try:
            await self.callback(*ctx.args, **ctx.kwargs)
        except Exception:
            ctx.command_failed = True
            raise
        finally:
            if call_hooks:
                await self.call_after_hooks()

    def error(self, coro: Awaitable) -> Awaitable:
        """A decorator that registers a coroutine as a local error handler.

        A local error handler is an :func:`.event_command_error` event limited
        to a single command.

        Command error handlers are raised in a specific order. Returning
        ``False`` in any of them will invoke the next handler in the chain. If
        there are no handlers left to call, then the error is printed to
        stderr (console).

        The order goes as follows:
        1. The local command error handler is called. (Handler specified by
        this decorator.)
        2. The local cog command error handler is called.
        3. All :func:`.event_command_error()` handlers are called
        simultaneously. If any of them return False, then the error will
        be printed.

        Parameters
        ----------
        coro:
            The coroutine to register as the local error handler.

        Raises
        ------
        TypeError
            The coroutine passed is not actually a coroutine.
        """

        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('Error handler must be a coroutine')

        self.error_handler = coro
        return coro

    def before_invoke(self, coro: Awaitable) -> Awaitable:
        """A decorator that registers a coroutine as a pre-invoke hook.

        A pre-invoke hook is called directly before the command is
        called. This makes it a useful function to set up database
        connections or any type of set up required.

        This pre-invoke hook takes a sole parameter, a :class:`.Context`.
        See :meth:`.Bot.before_invoke` for more info.

        Parameters
        -----------
        coro:
            The coroutine to register as the pre-invoke hook.

        Raises
        -------
        TypeError
            The coroutine passed is not actually a coroutine.
        """

        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('Before invoke handler must be a coroutine')

        self._before_invoke = coro
        return coro

    def after_invoke(self, coro: Awaitable) -> Awaitable:
        """A decorator that registers a coroutine as a post-invoke hook.

        A post-invoke hook is called directly after the command is
        called. This makes it a useful function to clean-up database
        connections or any type of clean up required.

        This post-invoke hook takes a sole parameter, a :class:`.Context`.
        See :meth:`.Bot.after_invoke` for more info.

        Parameters
        -----------
        coro:
            The coroutine to register as the post-invoke hook.

        Raises
        -------
        TypeError
            The coroutine passed is not actually a coroutine.
        """

        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('After invoke handler must be a coroutine')

        self._after_invoke = coro
        return coro

    @property
    def cog_name(self) -> str:
        """:class:`str`: The name of the cog this command
        belongs to. ``None`` otherwise.
        """
        return type(self.cog).__cog_name__ if self.cog is not None else None

    @property
    def short_doc(self) -> str:
        """:class:`str`: Gets the "short" documentation of a command.

        By default, this is the :attr:`brief` attribute.
        If that lookup leads to an empty string then the first line of the
        :attr:`help` attribute is used instead.
        """
        if self.brief is not None:
            return self.brief

        if self.help is not None:
            return self.help.split('\n', 1)[0]

        return ''

    def _is_typing_optional(self, annotation: Any) -> bool:
        try:
            origin = annotation.__origin__
        except AttributeError:
            return False

        if origin is not Union:
            return False

        NoneType = type(None)
        return annotation.__args__[-1] is NoneType

    @property
    def signature(self) -> str:
        """:class:`str`: Returns a POSIX-like signature useful
        for help command output.
        """

        if self.usage is not None:
            return self.usage

        params = self.clean_params
        if not params:
            return ''

        result = []
        for name, param in params.items():
            greedy = isinstance(param.annotation, converters._Greedy)

            if param.default is not param.empty:
                if isinstance(param.default, str):
                    should_print = param.default
                else:
                    should_print = param.default is not None

                if should_print:
                    result.append(
                        '[%s=%s]' % (name, param.default) if not greedy else
                        '[%s=%s]...' % (name, param.default)
                    )
                    continue
                else:
                    result.append('[%s]' % name)

            elif param.kind == param.VAR_POSITIONAL:
                result.append('[%s...]' % name)
            elif greedy:
                result.append('[%s]...' % name)
            elif self._is_typing_optional(param.annotation):
                result.append('[%s]' % name)
            else:
                result.append('<%s>' % name)

        return ' '.join(result)

    async def can_run(self, ctx: Context) -> bool:
        """|coro|

        Checks if the command can be executed by checking all the predicates
        inside the :attr:`.checks` attribute. This also checks whether the
        command is disabled.

        Parameters
        -----------
        ctx: :class:`.Context`
            The ctx of the command currently being invoked.

        Raises
        -------
        :class:`CommandError`
            Any command error that was raised during a check call will be
            propagated by this function.

        Returns
        --------
        :class:`bool`
            A boolean indicating if the command can be invoked.
        """

        if not self.enabled:
            raise errors.DisabledCommand(
                '{0.name} command is disabled'.format(self)
            )

        original = ctx.command
        ctx.command = self

        try:
            if not await ctx.bot.can_run(ctx):
                raise errors.CheckFailure(
                    'The global check functions for command '
                    '{0.qualified_name} failed.'.format(self))

            cog = self.cog
            if cog is not None:
                local_check = Cog._get_overridden_method(cog.cog_check)
                if local_check is not None:
                    if asyncio.iscoroutinefunction(local_check):
                        ret = await local_check(ctx)
                    else:
                        ret = local_check(ctx)

                    if not ret:
                        return False

            predicates = self.checks
            if not predicates:
                return True

            for func in predicates:
                if asyncio.iscoroutinefunction(func):
                    res = await func(ctx)
                else:
                    res = func(ctx)

                if not res:
                    return False
            return res

        finally:
            ctx.command = original


class GroupMixin:
    """A mixin that implements common functionality for classes that behave
    similar to :class:`.Group` and are allowed to register commands.

    Attributes
    -----------
    all_commands: :class:`dict`
        A mapping of command name to :class:`.Command` or subclass
        objects.
    case_insensitive: Optional[:class:`bool`]
        The passed value telling if the commands should be case
        insensitive. Defaults to ``None``. Use
        :attr:`~.GroupMixin.qualified_case_insensitive` to check if the command
        truly is case insensitive.
    """

    def __init__(self, *args: list, **kwargs: dict) -> None:
        super().__init__(*args, **kwargs)
        case_ins = kwargs.get('case_insensitive')
        self.all_commands = _CaseInsensitiveDict() if case_ins else {}
        self.case_insensitive = case_ins

    @property
    def qualified_case_insensitive(self) -> Optional[bool]:
        """:class:`bool`: The qualified case insensitive. This means that the
        it will never return ``None`` as it checks inherited values. Could
        be ``None`` but only if the bot is not ultimately registered to the bot
        yet.
        """
        if self.case_insensitive is not None:
            return self.case_insensitive

        parent = getattr(self, 'parent', None)
        if parent is not None:
            return self.parent.case_insensitive
        return self.instance.case_insensitive

    @property
    def commands(self) -> Set[Command]:
        """Set[:class:`.Command`]: A unique set of commands without aliases
        that are registered.
        """
        return set(self.all_commands.values())

    def recursively_remove_all_commands(self) -> None:
        for command in self.all_commands.copy().values():
            if isinstance(command, GroupMixin):
                command.recursively_remove_all_commands()
            self.remove_command(command.name)

    def recursively_make_case_insensitive(self) -> None:
        new_commands = _CaseInsensitiveDict()
        to_make = []
        for key, value in self.all_commands.items():
            new_commands[key] = value

            if isinstance(value, GroupMixin):
                to_make.append(value)

        self.all_commands = new_commands
        for mixin in to_make:
            mixin.recursively_make_case_insensitive()

    def add_command(self, command: Command) -> None:
        """Adds a :class:`.Command` or its subclasses into the internal list
        of commands.

        This is usually not called, instead the :meth:`~.GroupMixin.command` or
        :meth:`~.GroupMixin.group` shortcut decorators are used instead.

        Parameters
        -----------
        command: :class:`Command`
            The command to add.

        Raises
        -------
        :exc:`.ClientException`
            If the command is already registered.
        TypeError
            If the command passed is not a subclass of :class:`.Command`.
        """
        if not isinstance(command, Command):
            raise TypeError(
                'Command passed must be a subclassed instance of Command'
            )

        if isinstance(self, Command):
            command.parent = self

        if command.name in self.all_commands:
            raise errors.CommandError(
                'Command {0.name} is already registered.'.format(command)
            )

        if isinstance(command, GroupMixin):
            if command.case_insensitive is None:
                if command.qualified_case_insensitive:
                    command.recursively_make_case_insensitive()

        self.all_commands[command.name] = command
        for alias in command.aliases:
            if alias in self.all_commands:
                raise errors.CommandError(
                    'The alias {} is already an existing command or '
                    'alias.'.format(alias)
                )

            self.all_commands[alias] = command

    def remove_command(self, name: str) -> Optional[Command]:
        """Remove a :class:`.Command` or subclasses from the internal list
        of commands.

        This could also be used as a way to remove aliases.

        Parameters
        ----------
        name: :class:`str`
            The name of the command to remove.

        Returns
        --------
        :class:`.Command` or subclass
            The command that was removed. If the name is not valid then
            `None` is returned instead.
        """

        command = self.all_commands.pop(name, None)

        if command is None:
            return None

        if name in command.aliases:
            return command

        for alias in command.aliases:
            self.all_commands.pop(alias, None)

        return command

    def walk_commands(self) -> Iterable:
        """An iterator that recursively walks through all commands
        and subcommands.
        """

        for command in self.commands:
            yield command

            if isinstance(command, GroupMixin):
                yield from command.walk_commands()

    def get_command(self, name: str) -> Optional[Command]:
        """Get a :class:`.Command` or subclasses from the internal list
        of commands.

        This could also be used as a way to get aliases.

        The name could be fully qualified (e.g. ``'foo bar'``) will get
        the subcommand ``bar`` of the group command ``foo``. If a
        subcommand is not found then ``None`` is returned just as usual.

        Parameters
        -----------
        name: :class:`str`
            The name of the command to get.

        Returns
        --------
        :class:`Command` or subclass
            The command that was requested. If not found, returns ``None``.
        """

        if ' ' not in name:
            return self.all_commands.get(name)

        names = name.split()
        obj = self.all_commands.get(names[0])
        if not isinstance(obj, GroupMixin):
            return obj

        for name in names[1:]:
            try:
                obj = obj.all_commands[name]
            except (AttributeError, KeyError):
                return None

        return obj

    def command(self, *args: list, **kwargs: dict) -> callable:
        """A shortcut decorator that invokes :func:`.command` and adds
        it to the internal command list via :meth:`~.GroupMixin.add_command`.
        """

        def decorator(func):
            kwargs.setdefault('parent', self)
            result = command(*args, **kwargs)(func)
            self.add_command(result)
            return result

        return decorator

    def group(self, *args: list, **kwargs: dict) -> callable:
        """A shortcut decorator that invokes :func:`.group` and adds it to
        the internal command list via :meth:`~.GroupMixin.add_command`.
        """

        def decorator(func):
            kwargs.setdefault('parent', self)
            result = group(*args, **kwargs)(func)
            self.add_command(result)
            return result

        return decorator


class Group(GroupMixin, Command):
    """A class that implements a grouping protocol for commands to be
    executed as subcommands.

    This class is a subclass of :class:`.Command` and thus all options
    valid in :class:`.Command` are valid in here as well.

    Attributes
    -----------
    invoke_without_command: Optional[:class:`bool`]
        Indicates if the group callback should begin parsing and
        invocation only if no subcommand was found. Useful for
        making it an error handling function to tell the user that
        no subcommand was found or to have different functionality
        in case no subcommand was found. If this is ``False``, then
        the group callback will always be invoked first. This means
        that the checks and the parsing dictated by its parameters
        will be executed. Defaults to ``False``.
    case_insensitive: Optional[:class:`bool`]
        Indicates if the group's commands should be case insensitive.
        Defaults to ``None`` which means it inherits the parents or bots
        value.
    """

    def __init__(self, *args: list, **attrs: dict) -> None:
        self.invoke_without_command = attrs.pop(
            'invoke_without_command',
            False
        )
        super().__init__(*args, **attrs)

    async def invoke(self, ctx: Context) -> None:
        ctx.invoked_subcommand = None
        early_invoke = not self.invoke_without_command
        if early_invoke:
            await self.prepare(ctx)

        view = ctx.view
        previous = view.index
        view.skip_ws()
        trigger = view.get_word()

        if trigger:
            ctx.subcommand_passed = trigger
            ctx.invoked_subcommand = self.all_commands.get(trigger, None)

        if early_invoke:
            injected = hooked_wrapped_callback(self, ctx, self.callback)
            await injected(*ctx.args, **ctx.kwargs)

        if trigger and ctx.invoked_subcommand:
            ctx.invoked_with = trigger
            await ctx.invoked_subcommand.invoke(ctx)
        elif not early_invoke:
            # undo the trigger parsing
            view.index = previous
            view.previous = previous
            await super().invoke(ctx)

    async def reinvoke(self, ctx: Context, *,
                       call_hooks: bool = False) -> None:
        ctx.invoked_subcommand = None
        early_invoke = not self.invoke_without_command
        if early_invoke:
            ctx.command = self
            await self._parse_arguments(ctx)

            if call_hooks:
                await self.call_before_hooks(ctx)

        view = ctx.view
        previous = view.index
        view.skip_ws()
        trigger = view.get_word()

        if trigger:
            ctx.subcommand_passed = trigger
            ctx.invoked_subcommand = self.all_commands.get(trigger, None)

        if early_invoke:
            try:
                await self.callback(*ctx.args, **ctx.kwargs)
            except Exception:
                ctx.command_failed = True
                raise
            finally:
                if call_hooks:
                    await self.call_after_hooks(ctx)

        if trigger and ctx.invoked_subcommand:
            ctx.invoked_with = trigger
            await ctx.invoked_subcommand.reinvoke(ctx, call_hooks=call_hooks)

        elif not early_invoke:
            view.index = previous
            view.previous = previous
            await super().reinvoke(ctx, call_hooks=call_hooks)


def command(name: Optional[str] = None,
            cls: Optional[Command] = None,
            **attrs: dict) -> callable:
    """A decorator that transforms a function into a :class:`.Command`
    or if called with :func:`.group`, :class:`.Group`.

    By default the ``help`` attribute is received automatically from the
    docstring of the function and is cleaned up with the use of
    ``inspect.cleandoc``. If the docstring is ``bytes``, then it is decoded
    into :class:`str` using utf-8 encoding.

    All checks added using the :func:`.check` & co. decorators are added into
    the function. There is no way to supply your own checks through this
    decorator.

    Parameters
    -----------
    name: :class:`str`
        The name to create the command with. By default this uses the
        function name unchanged.
    cls
        The class to construct with. By default this is :class:`.Command`.
        You usually do not change this.
    attrs
        Keyword arguments to pass into the construction of the class denoted
        by ``cls``.

    Raises
    -------
    TypeError
        If the function is not a coroutine or is already a command.
    """

    if cls is None:
        cls = Command

    def decorator(func):
        if isinstance(func, Command):
            raise TypeError('Callback is already a command.')
        return cls(func, name=name, **attrs)

    return decorator


def group(name: Optional[str] = None, **attrs: dict) -> callable:
    """A decorator that transforms a function into a :class:`.Group`.

    This is similar to the :func:`.command` decorator but the ``cls``
    parameter is set to :class:`Group` by default.
    """
    attrs.setdefault('cls', Group)
    return command(name=name, **attrs)


def check(predicate: callable) -> MaybeCoro:
    r"""A decorator that adds a check to the :class:`.Command` or its
    subclasses. These checks could be accessed via :attr:`.Command.checks`.

    These checks should be predicates that take in a single parameter taking
    a :class:`.Context`. If the check returns a ``False``\-like value then
    during invocation a :exc:`.CheckFailure` exception is raised and sent to
    the :func:`.event_command_error` event.

    If an exception should be thrown in the predicate then it should be a
    subclass of :exc:`.CommandError`. Any exception not subclassed from it
    will be propagated while those subclassed will be sent to
    :func:`.event_command_error`.

    A special attribute named ``predicate`` is bound to the value
    returned by this decorator to retrieve the predicate passed to the
    decorator.

    .. note::

        The function returned by ``predicate`` is **always** a coroutine,
        even if the original function was not a coroutine.

    Examples
    ---------

    Creating a basic check to see if the command invoker is you.

    .. code-block:: python3

        def check_if_it_is_me(ctx):
            return ctx.author.id == "8b6373cbbe7d452b8172b5ee67ad53fa"

        @bot.command()
        @commands.check(check_if_it_is_me)
        async def only_for_me(ctx):
            await ctx.send('I know you!')

    Transforming common checks into its own decorator:

    .. code-block:: python3

        def is_me():
            def predicate(ctx):
                return ctx.author.id == "8b6373cbbe7d452b8172b5ee67ad53fa"
            return commands.check(predicate)

        @bot.command()
        @is_me()
        async def only_me(ctx):
            await ctx.send('Only you!')

    Parameters
    -----------
    predicate: Callable[[:class:`Context`], :class:`bool`]
        The predicate to check if the command should be invoked.
    """  # noqa

    def decorator(func):
        if isinstance(func, Command):
            func.checks.append(predicate)
        else:
            if not hasattr(func, '__fnpy_commands_checks__'):
                func.__fnpy_commands_checks__ = []

            func.__fnpy_commands_checks__.append(predicate)

        return func

    if asyncio.iscoroutinefunction(predicate):
        decorator.predicate = predicate
    else:
        @functools.wraps(predicate)
        async def wrapper(ctx):
            return predicate(ctx)

        decorator.predicate = wrapper

    return decorator


def check_any(*checks: list) -> callable:
    r"""A :func:`check` that is added that checks if any of the checks passed
    will pass, i.e. using logical OR.

    If all checks fail then :exc:`.CheckAnyFailure` is raised to signal the
    failure. It inherits from :exc:`.CheckFailure`.

    .. note::

        The ``predicate`` attribute for this function **is** a coroutine.

    Parameters
    ------------
    \*checks: Callable[[:class:`Context`], :class:`bool`]
        An argument list of checks that have been decorated with
        the :func:`check` decorator.

    Raises
    -------
    TypeError
        A check passed has not been decorated with the :func:`check`
        decorator.

    Examples
    ---------

    Creating a basic check to see if it's the bot owner or
    the party leader:

    .. code-block:: python3

        def is_party_leader():
            def predicate(ctx):
                return ctx.party is not None and ctx.author.leader
            return commands.check(predicate)

        @bot.command()
        @commands.check_any(commands.is_owner(), is_party_leader())
        async def only_for_owners(ctx):
            await ctx.send('Hello mister owner!')
    """

    unwrapped = []
    for wrapped in checks:
        try:
            pred = wrapped.predicate
        except AttributeError:
            raise TypeError(
                '{0} must be wrapped by commands.check '
                'decorator'.format(wrapped)
            ) from None
        else:
            unwrapped.append(pred)

    async def predicate(ctx):
        errors = []
        for func in unwrapped:
            try:
                value = await func(ctx)
            except errors.CheckFailure as e:
                errors.append(e)
            else:
                if value:
                    return True

        raise errors.CheckAnyFailure(unwrapped, errors)

    return check(predicate)


def cooldown(rate: int,
             per: float,
             type: BucketType = BucketType.default) -> callable:
    """A decorator that adds a cooldown to a :class:`.Command`
    or its subclasses.

    A cooldown allows a command to only be used a specific amount
    of times in a specific time frame. These cooldowns can be based
    either on a per-user or global basis.
    Denoted by the third argument of ``type`` which must be of enum
    type :class:`.BucketType`.

    If a cooldown is triggered, then :exc:`.CommandOnCooldown` is triggered in
    :func:`.event_command_error` and the local error handler.

    A command can only have a single cooldown.

    Parameters
    ------------
    rate: :class:`int`
        The number of times a command can be used before triggering a cooldown.
    per: :class:`float`
        The amount of seconds to wait for a cooldown when it's been triggered.
    type: :class:`.BucketType`
        The type of cooldown to have.
    """

    def decorator(func):
        if isinstance(func, Command):
            func._buckets = CooldownMapping(Cooldown(rate, per, type))
        else:
            func.__fnpy_commands_cooldown__ = Cooldown(rate, per, type)
        return func
    return decorator


def max_concurrency(number: int, per: BucketType = BucketType.default, *,
                    wait: bool = False) -> callable:
    """A decorator that adds a maximum concurrency to a :class:`.Command` or
    its subclasses.

    This enables you to only allow a certain number of command invocations at
    the same time, for example if a command takes too long or if only one user
    can use it at a time. This differs from a cooldown in that there is no set
    waiting period or token bucket -- only a set number of people can run the
    command.

    Parameters
    ----------
    number: :class:`int`
        The maximum number of invocations of this command that can be running
        at the same time.
    per: :class:`.BucketType`
        The bucket that this concurrency is based on, e.g. ``BucketType.user``
        would allow it to be used up to ``number`` times per user.
    wait: :class:`bool`
        Whether the command should wait for the queue to be over. If this is
        set to ``False`` then instead of waiting until the command can run
        again, the command raises :exc:`.MaxConcurrencyReached` to its error
        handler. If this is set to ``True`` then the command waits until it
        can be executed.
    """

    def decorator(func):
        value = MaxConcurrency(number, per=per, wait=wait)
        if isinstance(func, Command):
            func._max_concurrency = value
        else:
            func.__fnpy_commands_max_concurrency__ = value
        return func
    return decorator


def before_invoke(coro: Awaitable) -> callable:
    """A decorator that registers a coroutine as a pre-invoke hook.

    This allows you to refer to one before invoke hook for several commands
    that do not have to be within the same cog.

    Example
    ---------

    .. code-block:: python3

        async def record_usage(ctx):
            print(ctx.author, 'used', ctx.command, 'at', ctx.message.created_at)

        @bot.command()
        @commands.before_invoke(record_usage)
        async def who(ctx): # Output: <User> used who at <Time>
            await ctx.send('i am a bot')

        class What(commands.Cog):

            @commands.before_invoke(record_usage)
            @commands.command()
            async def who(self, ctx): # Output: <User> used when at <Time>
                await ctx.send('and my name is {}'.format(ctx.bot.user.display_name))

            @commands.command()
            async def where(self, ctx): # Output: <Nothing>
                await ctx.send('on Fortnite')

            @commands.command()
            async def why(self, ctx): # Output: <Nothing>
                await ctx.send('because someone made me')

        bot.add_cog(What())
    """  # noqa

    def decorator(func):
        if isinstance(func, Command):
            func.before_invoke(coro)
        else:
            func.__fnpy_before_invoke__ = coro
        return func
    return decorator


def after_invoke(coro: Awaitable) -> callable:
    """A decorator that registers a coroutine as a post-invoke hook.

    This allows you to refer to one after invoke hook for several commands that
    do not have to be within the same cog.
    """

    def decorator(func):
        if isinstance(func, Command):
            func.after_invoke(coro)
        else:
            func.__fnpy_after_invoke__ = coro
        return func
    return decorator


def dm_only() -> callable:
    """A :func:`.check` that indicates this command must only be used in a
    DM context. Only private messages are allowed when
    using the command.

    This check raises a special exception, :exc:`.PrivateMessageOnly`
    that is inherited from :exc:`.CheckFailure`.
    """

    def predicate(ctx):
        if ctx.party is not None:
            raise errors.PrivateMessageOnly()
        return True

    return check(predicate)


def party_only() -> callable:
    """A :func:`.check` that indicates this command must only be used in a
    party context only. Basically, no private messages are allowed when
    using the command.

    This check raises a special exception, :exc:`.PartyMessageOnly`
    that is inherited from :exc:`.CheckFailure`.
    """

    def predicate(ctx):
        if ctx.party is None:
            raise errors.PartyMessageOnly()
        return True

    return check(predicate)


def is_owner() -> callable:
    """A :func:`.check` that checks if the person invoking this command is the
    owner of the bot.

    This is powered by :meth:`.Bot.is_owner`.

    This check raises a special exception, :exc:`.NotOwner` that is derived
    from :exc:`.CheckFailure`.
    """

    async def predicate(ctx):
        if not await ctx.bot.is_owner(ctx.author.id):
            raise errors.NotOwner('You do not own this bot.')
        return True

    return check(predicate)
